#include <PalmOS.h>
#include <resource_ids.h>
#include <autoresources.h>
#include <toast_timer.h>
#include <stddef.h>
#include <toast_preferences.h>

#define DEFAULT_TIME 150
#define TMRNAME_MAX 10
#define TMRFLAG_RUNNING 0x0001
#define TMRFLAG_SOUND_OFF 0x0002
#define TMRFLAG_SOUND_MIDI 0x0004
#define SETTIME_FORMNAME_MAX TMRNAME_MAX + 20

typedef struct MidiRecord_struct *MidiRecord;

static struct AppGlobalsType_struct
{
  FormPtr main_form;
  FormPtr settime_form;
  FormPtr authors_form;
  FormPtr copying_form;
  FormPtr preferences_form;
  Int16 current_timer;
  MemHandle settime_form_name;
  Int32 entered_time;
  Boolean digits_entered;
  DmOpenRef dbref;
  DmOpenRef preferences;
  char preset_labels[ToastTimerNTimers][9];
  /* Bit fields indicating a need to synchronize the display with the
     database */
  UInt16 update_buttons;
  UInt16 update_counter_field;
  UInt16 update_preset_field;
  UInt16 update_name_field;

  MidiRecord midi_sounds;
  char **sound_names;
} AppGlobals;

typedef struct AppGlobalsType_struct *AppGlobalsType;

struct TimerRecord_struct
{
  
  UInt32 value;		/* When running: Wall clock time when the timer ends
			   When paused:  Seconds left
			*/
  UInt32 preset_value;	/* Seconds to run for. If zero count up */
  UInt32 flags;
  char name[TMRNAME_MAX+1];
  char sound_db_name[dmDBNameLength];
  UInt32 sound_rec_id;
};

typedef struct TimerRecord_struct *TimerRecord;

static struct TimerRecord_struct default_timer = {10,10,0,"Toast","",0};

static DmOpenRef
GetAppDB()
{
  UInt16 db_version;
  DmOpenRef dbref;
  Err error = 0;
  DmSearchStateType search;
  UInt16 card;
  LocalID dbid;
  error = DmGetNextDatabaseByTypeCreator (true, &search, ToastTimerDBType,
					  ToastTimerCreator, false, &card,
					  &dbid);
  if (error != errNone) {
    UInt16 r;
    if (error != dmErrCantFind) {
      ErrDisplay("Failed to find DB");
      return 0;
    }

    error = DmCreateDatabase(0,ToastTimerDBName,ToastTimerCreator,ToastTimerDBType,false);
    if (error) {
      ErrDisplay("Failed to create DB");
      return 0;
    }
    error = DmGetNextDatabaseByTypeCreator (true, &search, ToastTimerDBType,
					    ToastTimerCreator, false, &card,
					    &dbid);
    if (error != errNone) {
      ErrDisplay("Failed to find new DB");
      return 0;
    }
    dbref = DmOpenDatabase(card,dbid, dmModeReadWrite);
    if (dbref == 0) {
      DmDeleteDatabase(0,dbid);
      ErrDisplay("Failed to open new DB");
      return 0;
    }
    for (r = 0; r < ToastTimerNTimers && error == errNone; r++) {
      MemPtr timer;
      MemHandle rec;
      UInt16 rec_index = dmMaxRecordIndex;
      rec = DmNewRecord (dbref,&rec_index,sizeof(struct TimerRecord_struct));
      if (rec == 0) {
	LocalID dbid;
	dbid = DmFindDatabase(0,ToastTimerDBName);
	if (dbid) DmDeleteDatabase(0,dbid);
	ErrDisplay("Create DB record");
	return 0;
      }
      timer = MemHandleLock(rec);
      DmWrite(timer,0,&default_timer,sizeof(default_timer));
      MemHandleUnlock(rec);
      DmReleaseRecord(dbref,rec_index,true);
    }
    db_version = DB_VERSION;
    DmSetDatabaseInfo (card, dbid, NULL, NULL, &db_version , NULL, NULL,
		       NULL, NULL, NULL, NULL, NULL, NULL);
  } else {
    dbref = DmOpenDatabase(card,dbid, dmModeReadWrite);
    if (dbref == 0) {
      ErrDisplay("Failed to open DB");
      return 0;
    }
  }
  DmDatabaseInfo(card, dbid, NULL, NULL, &db_version,
              NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
  if (db_version < DB_VERSION) {
    UInt16 r;
    for (r = 0; r < ToastTimerNTimers && error == 0; r++) {
      struct TimerRecord_struct timer;
      MemHandle rec;
      MemPtr mem;
      rec = DmResizeRecord (dbref, r, sizeof(struct TimerRecord_struct)); 
      if (rec == NULL) {
	ErrDisplay("Failed to extend record");
	return 0;
      }
      mem = MemHandleLock(rec);
      timer = *(TimerRecord)mem;

      if (db_version < 1) {
	timer.sound_db_name[0] = '\0';
	timer.sound_rec_id = 0;
      }
      DmWrite(mem,0,&timer,sizeof(struct TimerRecord_struct));
      MemHandleUnlock(rec);
      DmReleaseRecord(dbref,r,true);
    }
    db_version = DB_VERSION;
    DmSetDatabaseInfo (card, dbid, NULL, NULL, &db_version, NULL, NULL,
		       NULL, NULL, NULL, NULL, NULL, NULL);
  }
  return dbref;
}


static void
FieldSetText(FieldType* field, char *str,UInt16 length)
{
  
  MemHandle txt_hndl = FldGetTextHandle(field);
  FldSetTextHandle(field,NULL);
  if (txt_hndl == NULL) {
    txt_hndl = MemHandleNew(length + 1);
  }
  if (txt_hndl) {
    char *txt_buffer = MemHandleLock(txt_hndl);
    StrCopy(txt_buffer,str);
    MemHandleUnlock(txt_hndl);
    FldSetTextHandle(field,txt_hndl);
  }
}

static void
StringNCat(char **pos, char *str, UInt16 len)
{
  while(len > 0) {
    *(*pos)++ = *str++;
    len--;
  }
}

static void
StringCat(char **pos, char *str)
{
  while(*str != '\0') {
    *(*pos)++ = *str++;
  }
}

static MemHandle
ParamString(char *template, char *param0)
{
  char *buffer;
  char *ppos;
  UInt16 template_len = StrLen(template);
  UInt16 param0_len = StrLen(param0);
  MemHandle buffer_hndl = MemHandleNew (template_len + param0_len);
  if (!buffer_hndl) return NULL;
  buffer = MemHandleLock(buffer_hndl);
  ppos = StrStr(template,"^1");
  if (ppos == NULL) {
    StrCopy(buffer,template);
  } else {
    char *pos = buffer;
    StringNCat(&pos, template, ppos - template);
    StringCat(&pos, param0);
    StringCat(&pos, ppos + 2);
    *pos = '\0';
  }
  MemHandleUnlock(buffer_hndl);
  return buffer_hndl;
}

static struct TimerIDs {
  UInt16 name_field;
  UInt16 count_field;
  UInt16 preset_field;
  UInt16 run_button;
  UInt16 pause_button;
  UInt16 restart_button;
} timer_ids[] =
{
  {NAME_0,COUNTDOWN_0,PRESET_0,RUN_0,PAUSE_0,RESTART_0},
  {NAME_1,COUNTDOWN_1,PRESET_1,RUN_1,PAUSE_1,RESTART_1},
  {NAME_2,COUNTDOWN_2,PRESET_2,RUN_2,PAUSE_2,RESTART_2},
  {NAME_3,COUNTDOWN_3,PRESET_3,RUN_3,PAUSE_3,RESTART_3},
  {NAME_4,COUNTDOWN_4,PRESET_4,RUN_4,PAUSE_4,RESTART_4},
  {NAME_5,COUNTDOWN_5,PRESET_5,RUN_5,PAUSE_5,RESTART_5},
  {NAME_6,COUNTDOWN_6,PRESET_6,RUN_6,PAUSE_6,RESTART_6},
  {NAME_7,COUNTDOWN_7,PRESET_7,RUN_7,PAUSE_7,RESTART_7},
};

static char
ChopDigit(UInt32 *v, UInt16 base)
{
  UInt32 r = *v / base;
  char c = '0' + (*v - r * base);
  *v = r;
  return c;
}

static void
PrintTime(char *buffer, UInt32 value)
{
  buffer[8] = '\0';
  buffer[7] = ChopDigit(&value, 10);
  buffer[6] = ChopDigit(&value, 6);
  buffer[5] = ':';
  buffer[4] = ChopDigit(&value, 10);
  buffer[3] = ChopDigit(&value, 6);
  buffer[2] = ':';
  buffer[1] = ChopDigit(&value, 10);
  buffer[0] = ChopDigit(&value, 10);
}

static void
UpdateCounter(FormPtr form,const TimerRecord timer,
	      UInt16 index,Boolean redraw)
{
  UInt32 count;
  char count_buffer[9];
  FieldType *field;
  /* Countdown field */
  field = FrmGetObjectPtr(form,
			  FrmGetObjectIndex(form,
					    timer_ids[index].count_field));
  if (timer->flags & TMRFLAG_RUNNING) {
    UInt32 current_time = TimGetSeconds();
    if (timer->preset_value == 0) {
      count = current_time - timer->value;
    } else {
      if (current_time >= timer->value)
	count = 0;
      else
	count = timer->value - current_time;
    }
  } else {
    count = timer->value;
  }
  PrintTime(count_buffer,count);
  FieldSetText(field,count_buffer,9);
  if (redraw) {
    FldDrawField(field);
  }
}


static void
UpdateNameField(FormPtr form,TimerRecord timer, UInt16 index)
{
  FieldType *field;
  /* Name field */
  field = FrmGetObjectPtr(form,
			  FrmGetObjectIndex(form,
					    timer_ids[index].name_field));
  FieldSetText(field,timer->name,StrLen(timer->name) + 1);
}

static void
SaveNameFields()
{
  UInt16 i;
  for (i = 0; i < ToastTimerNTimers;i++) {
    FieldPtr field;
    field = FrmGetObjectPtr(AppGlobals.main_form,
			    FrmGetObjectIndex(AppGlobals.main_form,
					      timer_ids[i].name_field));
    
    if (FldDirty(field)) {
      char *txt = FldGetTextPtr(field);
      TimerRecord mem;
      MemHandle rec = DmGetRecord(AppGlobals.dbref,i);
      if (!rec) return;
      mem = MemHandleLock(rec);
      DmStrCopy(mem,offsetof(struct TimerRecord_struct,name),txt);
      MemHandleUnlock(rec);
      DmReleaseRecord (AppGlobals.dbref,i,true);
    }
  }
}

static void
UpdatePreset(AppGlobalsType globals,TimerRecord timer, UInt16 index,Boolean redraw)
{
   ControlType *control;
  /* Preset field */
  
  control = FrmGetObjectPtr(globals->main_form,
			    FrmGetObjectIndex(globals->main_form,
					      timer_ids[index].preset_field));
  PrintTime(globals->preset_labels[index],timer->preset_value);
  CtlSetLabel(control,globals->preset_labels[index]);
  if (redraw) {
    CtlDrawControl(control);
  }
}

static void
SetControlValue(FormPtr form,UInt16 objid, Int16 value, Boolean redraw)
{
  UInt16 objidx = FrmGetObjectIndex(form,objid);
  ControlPtr control = FrmGetObjectPtr(form, objidx);
  if (!redraw) CtlSetUsable(control,false);
  CtlSetValue(control,value);
  if (!redraw) CtlSetUsable(control,true);
}

static void
UpdateButtons(TimerRecord timer, UInt16 index,FormPtr main_form,Boolean redraw)
{
  struct TimerIDs *ids = &timer_ids[index];
  
  SetControlValue(main_form,ids->run_button,
		  (timer->flags & TMRFLAG_RUNNING) ? 1: 0, redraw);

  SetControlValue (main_form, ids->pause_button,
		   ((timer->flags & TMRFLAG_RUNNING) == 0
		    && timer->value != timer->preset_value) ? 1: 0, redraw);
  
  SetControlValue(main_form, ids->restart_button,
		  ((timer->flags & TMRFLAG_RUNNING) == 0
		   && timer->value == timer->preset_value) ? 1: 0, redraw);
}


static void
UpdateMainForm(AppGlobalsType globals,Boolean redraw)
{
  UInt16 all = (globals->update_name_field | globals->update_counter_field
		| globals->update_preset_field | globals->update_buttons);
  UInt16 r;
  for (r = 0; r < ToastTimerNTimers; r++) {
    UInt16 mask = 1 << r;
    if (all & mask) {
      MemHandle rec = DmQueryRecord(AppGlobals.dbref,r);
      if (rec != 0) {
	TimerRecord timer = MemHandleLock(rec);
	if (globals->update_name_field & mask) {
	  UpdateNameField(globals->main_form,timer,r);
	  globals->update_name_field &= ~mask;
	}
	
	if (globals->update_counter_field & mask) {
	  UpdateCounter(globals->main_form,timer,r,redraw);
	  globals->update_counter_field &= ~mask;
	}

	if (globals->update_preset_field & mask) {
	  UpdatePreset(globals,timer,r,redraw);
	  globals->update_preset_field &= ~mask;
	}

	if (globals->update_buttons & mask) {
	  UpdateButtons(timer,r,globals->main_form,redraw);
	  globals->update_buttons &= ~mask;
	}
	
	MemHandleUnlock(rec);
      }
    }
  }
}

void
QueueCounterUpdate(UInt16 index)
{
  AppGlobals.update_counter_field |= 1 << index;
}

void
QueuePresetUpdate(UInt16 index)
{
  AppGlobals.update_preset_field |= 1 << index;
}

void
QueueButtonUpdate(UInt16 index)
{
  AppGlobals.update_buttons |= 1 << index;
}

#define UPDATE_ALL ((1<<ToastTimerNTimers) - 1)
static void
SetupMainForm(FormPtr form)
{
  AppGlobals.update_name_field = UPDATE_ALL;
  AppGlobals.update_counter_field = UPDATE_ALL;
  AppGlobals.update_preset_field = UPDATE_ALL;
  AppGlobals.update_buttons = UPDATE_ALL;
  UpdateMainForm(&AppGlobals,false);
}

struct MidiRecord_struct
{
  MidiRecord next;
  char *name;
  char *dbname;
  UInt32 recid;
};

static char* 
StrDup(char *str)
{
  char *new = MemPtrNew(sizeof(char) * (StrLen(str) + 1));
  if (!new) return NULL;
  StrCopy(new,str);
  return new;
}

static void
AddMidiRecord(MidiRecord *list,
	      char *name, char *dbname,UInt32 recid)
{
  MidiRecord record = MemPtrNew(sizeof(struct MidiRecord_struct));
  if (!record) return;
  record->next = *list;
  *list = record;
  record->name = StrDup(name);
  record->dbname = StrDup(dbname);
  record->recid = recid;
}

static void
FreeMidiRecordList(MidiRecord list)
{
  while(list) {
    MidiRecord next = list->next;
    MemPtrFree(list->dbname);
    MemPtrFree(list->name);
    MemPtrFree(list);
    list = next;
  }
}

static MidiRecord
FindMidiRecords()
{
  MidiRecord list = NULL;
  DmSearchStateType search;
  UInt16 card;
  LocalID dbid;
  UInt32 creator;
  Err err;
  creator = ToastGetPrefUseSysMIDIOnly() ? sysFileCSystem : 0;
  err = DmGetNextDatabaseByTypeCreator(true,&search,sysFileTMidi,creator,
				       true,&card,&dbid);
  while(err == errNone) {
    DmOpenRef dbref = DmOpenDatabase(card,dbid,dmModeReadOnly);
    if (dbref) {
      UInt16 r;
      UInt32 recid;
      char dbname[dmDBNameLength];
      UInt16 nrecs = DmNumRecords(dbref);
      DmDatabaseInfo (card, dbid, dbname, NULL, NULL, NULL, NULL, NULL,
		      NULL, NULL, NULL, NULL, NULL) ;
      for (r = 0; r < nrecs; r++) {
	MemHandle rec = DmQueryRecord(dbref,r);
	DmRecordInfo (dbref, r, NULL, &recid, NULL);
	if (rec) {
	  SndMidiRecType *midirec = MemHandleLock(rec);
	  AddMidiRecord(&list, midirec->name, dbname, recid);
	  MemHandleUnlock(rec);
	}
      }
      DmCloseDatabase(dbref);
    }
    err = DmGetNextDatabaseByTypeCreator(false,&search,sysFileTMidi,creator,
					 true,&card,&dbid);
  }
  return list;
}

static void
SetupSetTimeForm(FormPtr form)
{
  FieldType *field;
  
  if (AppGlobals.current_timer >= 0) {
    UInt32 version;
    MemHandle rec;
    MemHandle template_hndl;
    MemHandle str_hndl;
    TimerRecord timer;
    char str[9];
    char *sound_label;
    ControlPtr snd_ctl;
    ListPtr snd_list;
    UInt32 list_length = 2;
    Int16 current_sound = 1;
    SaveNameFields();
    rec = DmQueryRecord(AppGlobals.dbref, AppGlobals.current_timer);
    timer = (TimerRecord)MemHandleLock(rec);
    PrintTime(str,timer->preset_value);
    field = FrmGetObjectPtr(form, FrmGetObjectIndex(form, SETTIME_TIME_ID));
    FieldSetText(field,str,StrLen(str) + 1);
    AppGlobals.entered_time = timer->preset_value;
    AppGlobals.digits_entered = false;

    template_hndl = DmGetResource('tSTR',SETTIME_TITLE_TEMPLATE_ID);
    if (template_hndl) {
      char *template = MemHandleLock(template_hndl);
      AppGlobals.settime_form_name = ParamString(template, timer->name);
      MemHandleUnlock(template_hndl);
      DmReleaseResource(template_hndl);
  }

    /* Keep the title locked while displaying the form */
    FrmSetTitle (form, (char*)MemHandleLock(AppGlobals.settime_form_name));

    snd_list = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
						       SETTIME_SOUND_LIST_ID));

    FtrGet(sysFtrCreator, sysFtrNumROMVersion, &version);
    if (version >= 0x03003000) {
      MidiRecord midirec = FindMidiRecords();
      AppGlobals.midi_sounds = midirec;
      while(midirec) {
	midirec = midirec->next;
	list_length++;
      }
    }
    AppGlobals.sound_names = MemPtrNew(sizeof(char*) * list_length);
    
    str_hndl = DmGetResource('tSTR',SETTIME_SOUND_OFF_STRING_ID);
    AppGlobals.sound_names[0] = MemHandleLock(str_hndl);
    str_hndl = DmGetResource('tSTR',SETTIME_SOUND_SYSTEM_STRING_ID);
    AppGlobals.sound_names[1] = MemHandleLock(str_hndl);
    
    if (version >= 0x03003000) {
      UInt16 s = 2;
      MidiRecord midirec = AppGlobals.midi_sounds;
      while(midirec) {
	if (midirec->recid == timer->sound_rec_id
	    && StrCompare(midirec->dbname,timer->sound_db_name) == 0) {
	  current_sound = s;
	}

	AppGlobals.sound_names[s++] = midirec->name;
	midirec = midirec->next;
      }
    }
    
    if (timer->flags & TMRFLAG_SOUND_OFF) {
      current_sound = 0;
    } else if (!(timer->flags & TMRFLAG_SOUND_MIDI)) {
      current_sound = 1;
    }
    sound_label = AppGlobals.sound_names[current_sound];

    snd_ctl =FrmGetObjectPtr(form, FrmGetObjectIndex(form, SETTIME_SOUND_ID));
    CtlSetLabel(snd_ctl,sound_label);

    LstSetListChoices (snd_list, AppGlobals.sound_names, list_length);
    LstSetSelection(snd_list,current_sound);
    LstSetTopItem (snd_list, (current_sound < 2)? 0 : (current_sound - 1));
    MemHandleUnlock(rec);
  }
}
static void
CleanupSetTimeForm()
{
  MemHandle hndl;
  FreeMidiRecordList(AppGlobals.midi_sounds);
  AppGlobals.midi_sounds = NULL;

  hndl = MemPtrRecoverHandle(AppGlobals.sound_names[0]);
  MemHandleUnlock(hndl);
  DmReleaseResource(hndl);
  
  hndl = MemPtrRecoverHandle(AppGlobals.sound_names[1]);
  MemHandleUnlock(hndl);
  DmReleaseResource(hndl);

  MemPtrFree(AppGlobals.sound_names);
  AppGlobals.sound_names = NULL;
  
  MemHandleUnlock(AppGlobals.settime_form_name);
  MemHandleFree(AppGlobals.settime_form_name);
}

static void
GetTimer(DmOpenRef dbref,TimerRecord timer,UInt16 index)
{
  MemHandle rec = DmQueryRecord(dbref,index);
  *timer = *(TimerRecord)MemHandleLock(rec);
  MemHandleUnlock(rec);
}

#if 0
static void
SetTimer(DmOpenRef dbref,TimerRecord timer,UInt16 index)
{
  MemHandle rec = DmGetRecord(dbref,index);
  TimerRecord mem = MemHandleLock(rec);
  DmWrite(mem,0,timer,sizeof(struct TimerRecord_struct));
  MemHandleUnlock(rec);
  DmReleaseRecord(dbref,index,true);
}
#endif

/* Returns the timer with the lowest end time or -1 if
   none is running
*/
static Int16
CheckAlarm(DmOpenRef dbref,UInt32 *whenp)
{
  UInt32 index = ToastTimerNTimers;
  UInt32 when = ~0;
  UInt16 r;
  
  for (r = 0; r < ToastTimerNTimers; r++) {
    TimerRecord timer;
    MemHandle rec = DmQueryRecord(dbref,r);
    if (rec == 0) return -1;
    timer = MemHandleLock(rec);
    if (timer->flags & TMRFLAG_RUNNING && timer->preset_value != 0) {
      if (when > timer->value) {
	when = timer->value;
	index = r;
      }
    }
    MemHandleUnlock(rec);
  }
  if (index < ToastTimerNTimers) {
    *whenp = when;
    return index;
  }
  return -1;
}

static void
PostAlarm(UInt32 when, Int16 index)
{
  UInt16 card_no;
  LocalID db_ID;
  DmSearchStateType search_info;
  Err err;
  err = DmGetNextDatabaseByTypeCreator (true, &search_info, 
					  sysFileTApplication,
					  ToastTimerCreator,
					  true, &card_no, &db_ID);
  if (err) return;
  AlmSetAlarm(card_no,db_ID, (Int32)index, when, false);
}

#define TMRSOUND_OFF 0
#define TMRSOUND_SYSTEM 1

static void
ChangeSound(DmOpenRef dbref, UInt16 index, char *dbname, UInt32 recid)
{
  MemHandle rec = DmGetRecord(dbref,index);
  TimerRecord mem = MemHandleLock(rec);
  UInt32 flags = mem->flags;
  flags &= ~(TMRFLAG_SOUND_OFF | TMRFLAG_SOUND_MIDI);
  if (dbname) {
    DmStrCopy(mem,offsetof(struct TimerRecord_struct,sound_db_name),dbname);
    DmWrite(mem,offsetof(struct TimerRecord_struct,sound_rec_id),
	    &recid,sizeof(UInt32));
    flags |= TMRFLAG_SOUND_MIDI;
  } else if (recid == TMRSOUND_OFF) {
    flags |= TMRFLAG_SOUND_OFF;
  }
  DmWrite(mem,offsetof(struct TimerRecord_struct,flags),
	    &flags,sizeof(UInt32));
  MemHandleUnlock(rec);
  DmReleaseRecord(dbref,index,true);
}

#define TMRCMD_START 0x01
#define TMRCMD_PAUSE 0x02
#define TMRCMD_RESET 0x04
#define TMRCMD_SETPRESET 0x08

static void
ChangeTimer(DmOpenRef dbref, UInt16 index,UInt16 cmd, 
	    TimerRecord timer, UInt32 preset)
{
  Int16 next_index;
  UInt32 next_alarm;
  UInt32 current_time;
  MemHandle rec = DmGetRecord(dbref,index);
  TimerRecord mem = MemHandleLock(rec);
  *timer = *mem;
  if (TMRCMD_SETPRESET & cmd) {
    timer->preset_value = preset;
  }
  if (TMRCMD_RESET & cmd) {
    timer->value = timer->preset_value;
    timer->flags &= ~TMRFLAG_RUNNING;
  }
  current_time = TimGetSeconds();
  if (TMRCMD_START & cmd) {
    if ((timer->flags & TMRFLAG_RUNNING) == 0) {
      timer->flags |= TMRFLAG_RUNNING;
      if (timer->preset_value == 0) {
	timer->value = current_time - timer->value;
      } else {
	timer->value += current_time;
      }
    }
  } else if (TMRCMD_PAUSE & cmd) {
    if (timer->flags & TMRFLAG_RUNNING) {
      timer->flags &= ~TMRFLAG_RUNNING;
      if (timer->preset_value == 0) {
	timer->value = current_time - timer->value;
      } else {
	if (current_time >= timer->value)
	  timer->value = 0;
	else
	  timer->value -= current_time;
      }
    }
  }
  DmWrite(mem,0,timer,offsetof(struct TimerRecord_struct,name));
  MemHandleUnlock(rec);
  DmReleaseRecord(dbref,index,true);

  next_index = CheckAlarm(dbref,&next_alarm);
  if (next_index < 0) {
    PostAlarm(0,0);
  } else {
    if (next_alarm <= current_time) {
      next_alarm = current_time + 1;
    }
    PostAlarm(next_alarm,next_index);
  }  
}


static UInt16
GetAlarmVolume()
{
  UInt16 volume;
  UInt32 version;
  FtrGet(sysFtrCreator, sysFtrNumROMVersion, &version);
  if (version >= 0x03003000) { /* 3.X */
    volume = PrefGetPreference(prefAlarmSoundVolume);
  } else if (version >= 0x02003000) { /* 2.X */
    volume = (PrefGetPreference(prefAlarmSoundLevelV20) == slOn) ?sndMaxAmp :0;
  } else { /* 1.X */
    /* Is this the way to do it? */
    SystemPreferencesTypeV10 prefs;
    PrefGetPreferences((SystemPreferencesPtr)&prefs);
    volume = (prefs.alarmSoundLevel == slOn) ? sndMaxAmp : 0;
  }
  return volume;
}


static void
PlayAlarmSound(char *sound_db,UInt32 sound_rec_id)
{
/*    SndCommandType cmd; */
  if (sound_db) {
    LocalID dbid = DmFindDatabase(0,sound_db);
    if (dbid) {
      DmOpenRef dbref = DmOpenDatabase(0,dbid,dmModeReadOnly);
      if (dbref) {
	Err err;
	SndSmfOptionsType options;
	UInt8 *smf_data;
	SndMidiRecType *midirec;
	UInt16 index;
	MemHandle rec;
	DmFindRecordByID (dbref, sound_rec_id, &index);
	rec = DmQueryRecord (dbref, index);
	midirec = MemHandleLock(rec);
	smf_data = (UInt8*)midirec; /* + midirec->hdr.bDataOffset;*/
	options.dwStartMilliSec = 0;
	options.dwEndMilliSec = sndSmfPlayAllMilliSec;
	options.interruptible = true;
	options.amplitude = GetAlarmVolume();
	options.reserved = 0;
	err = SndPlaySmf(NULL,sndSmfCmdPlay,smf_data,&options,NULL,NULL,true);
	if (err != errNone && err != sndErrInterrupted) {
	  ErrDisplay("Failed to play MIDI sound");
	}
	MemHandleUnlock(rec);
	DmCloseDatabase(dbref);
      }
    }
  } else {
    SndPlaySystemSound(sndAlarm);
  /*
  cmd.cmd = sndCmdFreqDurationAmp;
  cmd.param1 = 2000;
  cmd.param2 = 150;
  cmd.param3 = volume;
  SndDoCmd (NULL, &cmd,0);
  cmd.param1 = 1500;
  cmd.param2 = 150;
  cmd.param3 = volume;
  SndDoCmd (NULL, &cmd,0);
  */
  }
}

static UInt16
TicksPerSecond()
{
  UInt32 romVersion;
  FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
  return (romVersion >= 0x2000000) ? SysTicksPerSecond() : 100;
}

Boolean
AlarmFormHandler(EventType *event)
{
  if (event->eType == ctlSelectEvent) {
    FrmReturnToForm(0);
    return true;
  }
  return false;
}



static void
DisplayAlarm(DmOpenRef dbref,UInt16 index,AppGlobalsType globals)
{
  FieldPtr field;
  UInt32 next_interval = TimGetSeconds();
  UInt32 ticks_per_second = TicksPerSecond();
  UInt32 repeats = ToastGetPrefRepeats();
  UInt32 interval = ToastGetPrefRepeatInterval();
  EventType event;
  FormPtr form;
  MemHandle template_hndl;
  struct TimerRecord_struct timer;
  FormPtr current_form;

  current_form = FrmGetActiveForm();
  if (current_form) FrmSetFocus(current_form,noFocus);

  GetTimer(dbref,&timer,index);
  form = FrmInitForm(TIMEOUT_ALERT_FORM_ID);

  field = FrmGetObjectPtr(form,
			  FrmGetObjectIndex(form, TIMEOUT_ALERT_MESSAGE_ID));

  template_hndl = DmGetResource('tSTR',TIMEOUT_ALERT_TEMPLATE_ID);
  if (template_hndl) {
    char *template = MemHandleLock(template_hndl);
    MemHandle txt_buffer = ParamString(template, timer.name);
    MemHandleUnlock(template_hndl);
    FldSetTextHandle(field,txt_buffer);

    DmReleaseResource(template_hndl);
  }
  
  FrmSetEventHandler(form,AlarmFormHandler);

  
  FrmDrawForm(form);
  FrmSetActiveForm(form);
  
  ChangeTimer(dbref,index,TMRCMD_RESET,&timer,0);

  do {
    UInt32 current_time = TimGetSeconds();
    UInt32 wait_ticks;
    event.eType = nilEvent;
    if (next_interval > current_time) {
      wait_ticks = ticks_per_second * (next_interval - current_time);
    } else {
      wait_ticks = 0;
    }
    EvtGetEvent (&event, wait_ticks);
    current_time = TimGetSeconds();
    if (current_time >= next_interval) {
      if (!(timer.flags & TMRFLAG_SOUND_OFF)) {
	if (repeats > 0) {
	  if (timer.flags & TMRFLAG_SOUND_MIDI) {
	    PlayAlarmSound(timer.sound_db_name,timer.sound_rec_id);
	  } else {
	    PlayAlarmSound(NULL,0);
	  }
	  repeats--;
	}
      }
      next_interval = interval + TimGetSeconds();
    }
    if (event.eType != nilEvent) {
      if (!SysHandleEvent (&event)) {
	FrmDispatchEvent(&event);
      }
    }
  } while (event.eType != ctlSelectEvent);
  
  if (globals != NULL) {
    QueueButtonUpdate(index);
    QueueCounterUpdate(index);
  }
}

#if 0
static void
DisplayPendingAlarms(DmOpenRef dbref,AppGlobalsType globals)
{
  UInt32 when = 0;
  Int16 index;
  while(true) {
    index = CheckAlarm(dbref,&when);
    if (index < 0) break;
    if (when > TimGetSeconds()) break;
    DisplayAlarm(dbref,index,globals);
  }
}
#endif

static Boolean
AuthorsFormHandler(EventType *event)
{
  if (event->eType == ctlSelectEvent) {
    FrmGotoForm(MAINFORM_ID);
  } else if (event->eType == frmCloseEvent) {
    AppGlobals.authors_form = NULL;
  }
  return false;
}


static void
SyncScrollbarToField(FieldPtr field, ScrollBarPtr scrollbar)
{
  UInt16 max,page_size,value,text_height,field_height;
  
  FldGetScrollValues(field, &value, &text_height, &field_height);
  if (text_height > field_height) 
    max = text_height - field_height;
  else
    max = value;

  page_size = field_height - 3;
  SclSetScrollBar(scrollbar,value,0,max,page_size);
}

static void
SetupCopyingForm()
{
  ScrollBarPtr scrollbar;
  MemHandle str_hndl;
  FieldPtr field;
  field = FrmGetObjectPtr(AppGlobals.copying_form,
			  FrmGetObjectIndex(AppGlobals.copying_form,
					    COPYING_FIELD_ID));
  str_hndl = DmGetResource('tSTR',COPYING_STRING_ID);
  if (str_hndl) {
    FldSetTextHandle(field,str_hndl);
  }
  scrollbar = FrmGetObjectPtr(AppGlobals.copying_form,
			      FrmGetObjectIndex(AppGlobals.copying_form,
						COPYING_SCROLLBAR_ID));
  SyncScrollbarToField(field,scrollbar);
}

static void
CleanupCopyingForm()
{
  MemHandle str_hndl;
  FieldPtr field;
  field = FrmGetObjectPtr(AppGlobals.copying_form,
			  FrmGetObjectIndex(AppGlobals.copying_form,
					    COPYING_FIELD_ID));
  str_hndl = FldGetTextHandle(field);
  if (str_hndl) {
    FldSetTextHandle(field,NULL);
    DmReleaseResource(str_hndl);
  }
}



static Boolean
CopyingFormHandler(EventType *event)
{
  if (event->eType == ctlSelectEvent) {
    FrmGotoForm(MAINFORM_ID);
  } else if (event->eType == frmCloseEvent) {
    CleanupCopyingForm();
    AppGlobals.copying_form = NULL;
  } if (event->eType == fldChangedEvent) {
    FieldPtr field;
    ScrollBarPtr scrollbar;
    scrollbar = FrmGetObjectPtr(AppGlobals.copying_form,
				FrmGetObjectIndex(AppGlobals.copying_form,
						  COPYING_SCROLLBAR_ID));
    field = event->data.fldChanged.pField;
    SyncScrollbarToField(field,scrollbar);
    
  } else if (event->eType == sclRepeatEvent) {
    Int16 delta;
    WinDirectionType dir = winDown;
    FieldPtr field;
    field = FrmGetObjectPtr(AppGlobals.copying_form,
			    FrmGetObjectIndex(AppGlobals.copying_form,
					      COPYING_FIELD_ID));
    delta = event->data.sclRepeat.newValue - event->data.sclRepeat.value;
    if (delta < 0) {
      delta = -delta;
      dir = winUp;
    }
    FldScrollField(field,delta,dir);
  }
  return false;
}


static void
SetupPreferencesForm(FormPtr form)
{
  FieldPtr field;
  ControlPtr ctl;
  char buffer[7];
  field = FrmGetObjectPtr(form, FrmGetObjectIndex(form, REPEATS_FIELD_ID));
  StrPrintF(buffer,"%ld",ToastGetPrefRepeats());
  FieldSetText(field,buffer, StrLen(buffer));

  field = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
						  REPEAT_INTERVAL_FIELD_ID));
  StrPrintF(buffer,"%ld",ToastGetPrefRepeatInterval());
  FieldSetText(field,buffer, StrLen(buffer));

  ctl = FrmGetObjectPtr(form,
			FrmGetObjectIndex(form, USE_SYS_MIDI_ONLY_CHECK_ID));
  CtlSetValue(ctl, ToastGetPrefUseSysMIDIOnly() ? 1 : 0);
}


static Boolean
PreferencesFormHandler(EventType *event)
{
  if (event->eType == ctlSelectEvent) {
    switch(event->data.ctlSelect.controlID) {
    case PREFS_DONE_ID:
      FrmGotoForm(MAINFORM_ID);
      return true;
    }
  } else if (event->eType == frmCloseEvent) {
    FieldPtr field;
    ControlPtr ctl;
    char *txt;
    field = FrmGetObjectPtr(AppGlobals.preferences_form,
			    FrmGetObjectIndex(AppGlobals.preferences_form,
					      REPEATS_FIELD_ID));
    txt = FldGetTextPtr(field);
    ToastSetPrefRepeats(AppGlobals.preferences, StrAToI(txt));

    field = FrmGetObjectPtr(AppGlobals.preferences_form,
			    FrmGetObjectIndex(AppGlobals.preferences_form,
					      REPEAT_INTERVAL_FIELD_ID));
    txt = FldGetTextPtr(field);
    ToastSetPrefRepeatInterval(AppGlobals.preferences, StrAToI(txt));
    
    ctl = FrmGetObjectPtr(AppGlobals.preferences_form,
			  FrmGetObjectIndex(AppGlobals.preferences_form,
					    USE_SYS_MIDI_ONLY_CHECK_ID));
    ToastSetPrefUseSysMIDIOnly(AppGlobals.preferences,CtlGetValue(ctl));
    
    AppGlobals.preferences_form = NULL;
  }
  return false;
}

static Boolean
MainFormHandler(EventType *event)
{
  if (event->eType == ctlSelectEvent) {
    UInt16 r;
    UInt16 id = event->data.ctlSelect.controlID;
    for (r = 0; r < ToastTimerNTimers; r++) {
      if (id == timer_ids[r].run_button) {
	struct TimerRecord_struct timer;
	ChangeTimer(AppGlobals.dbref,r,TMRCMD_START,&timer,0);
      } else if (id == timer_ids[r].pause_button) {
	struct TimerRecord_struct timer;
	ChangeTimer(AppGlobals.dbref,r,TMRCMD_PAUSE,&timer,0);
	QueueCounterUpdate(r);
      } else if (id == timer_ids[r].restart_button) {
	struct TimerRecord_struct timer;
	ChangeTimer(AppGlobals.dbref,r,TMRCMD_RESET,&timer,0);
	QueueCounterUpdate(r);
      } else if (id == timer_ids[r].preset_field) {
	AppGlobals.current_timer = r;
	/* Clear the cursor, otherwise it will screw up the popup */
	FrmSetFocus(AppGlobals.main_form,noFocus);
	FrmPopupForm(SETTIME_FORM_ID);
      }
    }
  } else if (event->eType == frmSaveEvent) {
    SaveNameFields();
  } else if (event->eType == menuEvent) {
    switch(event->data.menu.itemID) {
    case MAIN_MENU_AUTORS_ID:
      FrmGotoForm(AUTHORS_FORM_ID);
      break;
    case MAIN_MENU_COPYING_ID:
      FrmGotoForm(COPYING_FORM_ID);
      break;
    case MAIN_MENU_PREFERENCES_ID:
      FrmGotoForm(PREFERENCES_FORM_ID);
      break;
      
    }
  } else if (event->eType == frmCloseEvent) {
    AppGlobals.main_form = NULL;
  }
  return false;
}

/* Returns -1 if the string doesn't contain a legal time */
Int32
ParseTime(char *str)
{
  Int16 numbers[3];
  Int32 sum = 0;
  Int16 i = 0;
  while(*str != '\0' && i < 3) {
    if (*str >= '0' && *str <='9') {
      sum = sum * 10 + (*str -'0');
    } else if (*str == ':') {
      numbers[i++] = sum;
      sum = 0;
    } else {
      return -1;
    }
    str++;
  }
  if (i == 3) return -1;
  numbers[i++] = sum;
  if (i == 0) return -1;
  sum = numbers[i-1];
  if (sum >= 60) return -1;
  if (i >= 2) {
    if (numbers[i-2] >= 60) return -1;
    sum += numbers[i-2] * 60;
    if (i >= 3) {
      if (numbers[i-3] >= 100) return -1;
      sum += (UInt32)numbers[i-3] * 3600;
    }
  }
  return sum;
}

static void
FrmHideControl(FormType *formP, UInt16 objIndex)
{
  UInt32 version;
  FrmHideObject(formP,objIndex);
  FtrGet(sysFtrCreator, sysFtrNumROMVersion, &version);
  if (version < 0x03203000) {
    ControlPtr ctrl = (ControlPtr)FrmGetObjectPtr(formP,objIndex);
    ctrl->attr.usable = false;
  }
}

		   
#define KEYCASE(k) case SETTIME_ ## k : append = #k ; break
Boolean
SetTimeFormHandler(EventType *event)
{
  if (event->eType == ctlSelectEvent) {
    UInt16 id = event->data.ctlSelect.controlID;
    Boolean popdown = false;
    char *append = NULL;
    switch(id) {
    case SETTIME_OK_ID:
    case SETTIME_START_ID:
      if (AppGlobals.entered_time >= 0) {
	ListPtr snd_list;
	Int16 selected_snd;
	struct TimerRecord_struct timer;
	UInt16 cmd = TMRCMD_SETPRESET | TMRCMD_RESET;
	if (id == SETTIME_START_ID) cmd |= TMRCMD_START;
	ChangeTimer(AppGlobals.dbref,AppGlobals.current_timer,cmd,&timer,AppGlobals.entered_time);

	snd_list = FrmGetObjectPtr(AppGlobals.settime_form,
				   FrmGetObjectIndex(AppGlobals.settime_form,
						     SETTIME_SOUND_LIST_ID));

	selected_snd = LstGetSelection (snd_list);
	if (noListSelection != selected_snd) {
	  if (selected_snd == 0) {
	    ChangeSound(AppGlobals.dbref,AppGlobals.current_timer,
			NULL, TMRSOUND_OFF);
	    } else if (selected_snd == 1) {	/* System sound */
	      ChangeSound(AppGlobals.dbref,AppGlobals.current_timer,
			  NULL, TMRSOUND_SYSTEM);
	    } else {		/* MIDI sound */
	      MidiRecord info = AppGlobals.midi_sounds;
	      while(info && selected_snd > 2) {
		info = info->next;
		selected_snd--;
	      }
	      if (info) {
		ChangeSound(AppGlobals.dbref,AppGlobals.current_timer,
			    info->dbname,info->recid);
	      }
	    }
	}
	QueuePresetUpdate(AppGlobals.current_timer);
	QueueButtonUpdate(AppGlobals.current_timer);
	QueueCounterUpdate(AppGlobals.current_timer);
      }
    case SETTIME_CANCEL_ID:
      popdown = true;
      break;
      
      KEYCASE(0);
      KEYCASE(1);
      KEYCASE(2);
      KEYCASE(3);
      KEYCASE(4);
      KEYCASE(5);
      KEYCASE(6);
      KEYCASE(7);
      KEYCASE(8);
      KEYCASE(9);
      KEYCASE(00);
    case SETTIME_CLEAR_ID:
      append = "000000";
      break;
    }
    if (append) {
      Int32 prev_entered = AppGlobals.entered_time;
      FieldPtr field;
      MemHandle txth;
      char *txt;
      UInt16 i;
      UInt16 len = StrLen(append);
      const static int digit_indexes[] = {0,1,3,4,6,7};
      field = FrmGetObjectPtr(AppGlobals.settime_form,
			      FrmGetObjectIndex(AppGlobals.settime_form,
						SETTIME_TIME_ID));
      txth = FldGetTextHandle(field);
      FldSetTextHandle(field,NULL);
      txt = MemHandleLock(txth);
      if (AppGlobals.digits_entered) {
	for (i = 0; i < 6 - len; i++) {
	  txt[digit_indexes[i]] = txt[digit_indexes[i+len]];
	}
      } else {
	StrCopy(txt,"00:00:00");
      }
      for(i = 1; i <= len; i++) {
	txt[digit_indexes[6 - i]] = append[len - i];
      }
      AppGlobals.entered_time = ParseTime(txt);

      MemHandleUnlock(txth);
      FldSetTextHandle(field,txth);
      FldDrawField(field);
      if (AppGlobals.entered_time >= 0 && prev_entered < 0) {
	FrmShowObject (AppGlobals.settime_form,
		       FrmGetObjectIndex(AppGlobals.settime_form,
					 SETTIME_OK_ID));
	FrmShowObject (AppGlobals.settime_form,
		       FrmGetObjectIndex(AppGlobals.settime_form,
					 SETTIME_START_ID)); 
      } else if (AppGlobals.entered_time < 0 && prev_entered >= 0) {
	FrmHideControl(AppGlobals.settime_form,
		       FrmGetObjectIndex(AppGlobals.settime_form,
					 SETTIME_OK_ID)); 
	FrmHideControl(AppGlobals.settime_form,
		       FrmGetObjectIndex(AppGlobals.settime_form,
					 SETTIME_START_ID));
      }
      AppGlobals.digits_entered = true;
    }
    if (popdown) {
      FrmReturnToForm(MAINFORM_ID);
      CleanupSetTimeForm();
      AppGlobals.settime_form = NULL;
      FrmDrawForm(AppGlobals.main_form);
    }
    return popdown || append != NULL;
  } if (event->eType == popSelectEvent) {
    Int16 selected = event->data.popSelect.selection;
    if (event->data.popSelect.listID == SETTIME_SOUND_LIST_ID) {
      if (selected >= 1) {
	if (selected == 1) {	/* System sound */
	  PlayAlarmSound(NULL,0);
	} else {		/* MIDI sound */
	  MidiRecord info = AppGlobals.midi_sounds;
	  while(info && selected > 2) {
	    info = info->next;
	    selected--;
	  }
	  if (info) {
	    PlayAlarmSound(info->dbname,info->recid);
	  }
	}
      }
    }
  } else if (event->eType == frmCloseEvent) {
    AppGlobals.settime_form = NULL;
  }
  return false;
}

/* Returns true if any of the counters has reached zero */
static Boolean
UpdateCounters()
{
  UInt16 r;
  Boolean zero = false;
  UInt32 now = TimGetSeconds();
  for (r = 0; r < ToastTimerNTimers; r++) {
    struct TimerRecord_struct timer;
    GetTimer(AppGlobals.dbref,&timer,r);
    if (timer.flags & TMRFLAG_RUNNING) {
      QueueCounterUpdate(r);
      if (timer.value <= now) {
	zero = true;
      }
    }
  }
  return zero;
}

static Boolean
ApplicationHandleEvent (EventType *event)
{
  UInt16 form_id;
  if (event->eType == frmLoadEvent) {
    FormPtr *formp = NULL;
    form_id = event->data.frmLoad.formID;
    switch(form_id) {
    case MAINFORM_ID:
      formp = &AppGlobals.main_form;
      break;
    case SETTIME_FORM_ID:
      formp = &AppGlobals.settime_form;
      break;
    case AUTHORS_FORM_ID:
      formp = &AppGlobals.authors_form;
      break;
    case COPYING_FORM_ID:
      formp = &AppGlobals.copying_form;
      break;
    case PREFERENCES_FORM_ID:
      formp = &AppGlobals.preferences_form;
      break;
    }
    if (*formp == NULL) {
      *formp = FrmInitForm(form_id);
    }
    return true;
  } else if (event->eType == frmOpenEvent) {
    FormPtr form = NULL;
    form_id = event->data.frmOpen.formID;
    switch(form_id) {
    case MAINFORM_ID:
      {
	form = AppGlobals.main_form;
	SetupMainForm(form);
	FrmSetEventHandler(form,MainFormHandler);
      }
      break;
    case SETTIME_FORM_ID:
      {
	form = AppGlobals.settime_form;
	SetupSetTimeForm(form);
	FrmSetEventHandler(form,SetTimeFormHandler);
      }
      break;
    case AUTHORS_FORM_ID:
      {
	form = AppGlobals.authors_form;
	/* SetupAuthorsForm(form);*/
	FrmSetEventHandler(form,AuthorsFormHandler);
      }
      break;
    case COPYING_FORM_ID:
      {
	form = AppGlobals.copying_form;
	SetupCopyingForm(form);
	FrmSetEventHandler(form,CopyingFormHandler);
      }
      break;
    case PREFERENCES_FORM_ID:
      {
	form = AppGlobals.preferences_form;
	SetupPreferencesForm(form);
	FrmSetEventHandler(form,PreferencesFormHandler);
      }
      break;
    }
    FrmDrawForm(form);
    FrmSetActiveForm(form);
    return true;
  }
  
  return false;
}


static void
EventLoop (void)
{
  UInt16 error;
  EventType event;
  UInt16 wait_ticks = TicksPerSecond() / 10;
  UInt32 last_time = TimGetSeconds();
  do {
    UInt32 current_time;
    event.eType = nilEvent;
    EvtGetEvent (&event, wait_ticks);
    current_time = TimGetSeconds();
    if (current_time != last_time) {
      if (FrmGetActiveFormID() == MAINFORM_ID) {
	if (UpdateCounters()) {
	  SaveNameFields();
	  /*
	  PlayAlarmSound();
	  DisplayPendingAlarms(AppGlobals.dbref,AppGlobals.main_form);
	  */
	}
      }
      last_time = current_time;
      if (AppGlobals.main_form != NULL
	  && FrmGetActiveForm() == AppGlobals.main_form)
	UpdateMainForm(&AppGlobals, true);
    }
    if (event.eType != nilEvent) {
      if (!SysHandleEvent (&event)) {
	if (! MenuHandleEvent (NULL, &event, &error)) {
	  
	  if (! ApplicationHandleEvent (&event)) { 
	    
	    FrmDispatchEvent (&event); 
	  } 
	}
      }
      if (AppGlobals.main_form != NULL
	  && FrmGetActiveForm() == AppGlobals.main_form)
	UpdateMainForm(&AppGlobals, true);
    }
  } while (event.eType != appStopEvent);
}

static UInt16
StartApplication()
{

  AppGlobals.main_form = NULL;
  AppGlobals.settime_form = NULL;
  AppGlobals.authors_form = NULL;
  AppGlobals.copying_form = NULL;
  AppGlobals.current_timer = -1;
  AppGlobals.dbref = GetAppDB();
  if (AppGlobals.dbref == 0) {
    return dmErrCantOpen;
  }
  AppGlobals.preferences = ToastGetPreferences();
  AppGlobals.midi_sounds = NULL;
  AppGlobals.sound_names = NULL;
  return 0;
}

static UInt16
StopApplication()
{
  FrmSaveAllForms();
  FrmCloseAllForms();
  ToastReleasePreferences(AppGlobals.preferences);
  DmCloseDatabase(AppGlobals.dbref);
  AppGlobals.dbref = 0;
  return 0;
}

UInt32
PilotMain (UInt16 cmd, MemPtr cmd_arg, UInt16 launch_flags)
{
  UInt16 error;
  if (cmd == sysAppLaunchCmdNormalLaunch) {
    error = StartApplication ();
    if (error) return error;
    FrmGotoForm (MAINFORM_ID);
    EventLoop ();
    StopApplication ();
  } else if (cmd == sysAppLaunchCmdAlarmTriggered) {
    Int32 ref = ((SysAlarmTriggeredParamType*)cmd_arg)->ref;
    
    if (ref < 0) {
      if (ref < -1) PostAlarm(TimGetSeconds() + 2,ref + 1);
      ((SysAlarmTriggeredParamType*)cmd_arg)->purgeAlarm = true;
    } else {
      ((SysAlarmTriggeredParamType*)cmd_arg)->purgeAlarm = false;
    }
  } else if (cmd == sysAppLaunchCmdDisplayAlarm) {
    DmOpenRef dbref;
    DmOpenRef prefs = NULL;
    AppGlobalsType globals;
    if (launch_flags & sysAppLaunchFlagSubCall) {
      dbref = AppGlobals.dbref;
      globals = &AppGlobals;
    } else {
      dbref = GetAppDB();
      globals = NULL;
      prefs = ToastGetPreferences();
    }
    if (dbref != 0) {
      Int16 index = (Int32)((SysAlarmTriggeredParamType*)cmd_arg)->ref;

      PostAlarm(TimGetSeconds() ,-10);

      DisplayAlarm(dbref,index,globals);

 
      if (launch_flags & sysAppLaunchFlagSubCall) {
	if (AppGlobals.main_form != NULL
	    && FrmGetActiveForm() == AppGlobals.main_form)
	  UpdateMainForm(&AppGlobals, true);
      } else {
	DmCloseDatabase(dbref);
	ToastReleasePreferences(prefs);
      }
    }
  }
  
  return 0;
}
